/*This file is adapted from Cecile Ane's Seq-Gen-Cov, with the addition of the Galtier model, and modified to work with amino acids as well as nucleotides*/
/*24 Feb 2006: added a general model that contains both Huelsenbeck and Galtier*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "global.h"
#include "tree.h"
#include "evolve.h"
#include "model.h"
#include "nucmodels.h"
#include "aamodels.h"
#include "gamma.h"
#include "twister.h"
#include "covarion.h" 


/*#define DEBUG*/

int numTaxa, numSites, maxPartitions, numPartitions, fileFormat;
double covarionRate, ONfreq;
double CovRate[3];
double pcov; /*proportion of covarion sites in Galtier model and general model*/
double s01,s10,s11;/*switching rates under general model*/
int covarion;
char *onoff="-+";
double freqRate[MAX_RATE_CATS];



/* prototypes */

signed char SetHiddenState();
signed char SetRASState(signed char j);
signed char SetGeneralOnOff(signed char j);
void RandomHiddenStates(signed char *seq, int inNumSites);
void Huel_covarionSimulation(signed char *state, double *effectiveLength, double len);/*Huelsenbeck model, from Ane*/
void Galt_covarionSimulation(signed char *state, double *effectiveLength, double len);/*Galtier model, added by MS*/
void MutateHiddenStates(signed char *seq, double *effectiveLength, int inFromSite, int inNumSites, double len);
void EvolveHiddenStateNode(TNode *anc, TNode *des, int inFromSite, int inNumSites);
void WriteHiddenStatesNode(FILE *fv, TTree *tree, int *nodeNo, TNode *des, int inNumSites,TTree *nametree);
void WriteAncestralEffLengthNode(FILE *fv, TTree *tree, int *nodeNo, TNode *des, int inNumSites, TTree *nametree);


/* functions */


void CreateHiddenStates(TTree *tree, int inNumSites)
{
  /* Creates memory for the "hiddenStates" and the "effectiveLength0" pointers
     of all the nodes of the tree. 
  */

	TNode *P;
	
	P=tree->nodeList;
	while (P!=NULL) {
		if (P->hiddenStates != NULL)
			free(P->hiddenStates);
			
		P->hiddenStates=CAllocMem(inNumSites+1, "hiddenStates", "CreateHiddenStates", 0);

		if (P->effectiveLength0 != NULL)
		  free(P->effectiveLength0 );
		P->effectiveLength0=CAllocMem(sizeof(double)*(inNumSites+1),"hiddenLengths", "CreateHiddenStates", 0);
		P=P->next;
	}
}



signed char SetHiddenState()
{
	signed char j;
	double r;
	r=rndu();

	switch(covarion) {/* creates a state on/off at random according to the probability ONfreq. */
		case Huel:/*hidden states are 0,1*/
			j='\1';   		/* ON */
			if (r > ONfreq) j='\0'; /* OFF */
		break;
		case Galt: case General:/*hidden states are 1...g for covarion, -1...-g for RAS (and will be g+1...2g for general covarion OFF states)*/
			/*each state is equiprobable*/
			j=(signed char)(rndu()*numCats)+1;
		break;
	default:j='\0';/*shouldn't happen*/
	}
	return j;
}

signed char SetRASState(signed char j)
{
	/*proportion 1-pcov of sites have a constant gamma rate*/
	if (rndu()>pcov) {
		j=-j;/*negative state has rate fixed once, never draw again*/
	}
	return j;
}

signed char SetGeneralOnOff(signed char j)
{
	if (rndu()>s01/(s01+s10)) {/*proportion s01/(s01+s10) of covarion sites initially ON*/
		j+=numCats;/*states g+1...2g are covarion OFF states*/
	}
	return j;
}

void RandomHiddenStates(signed char *seq, int inNumSites)
{
/* Huelsenbeck model: creates a sequence of hidden states of length inNumSites,
   according to the probability ONfreq to be 'on'
  Galtier model: proportion pcov of sites have states 1...g equiprobable covarion, proportion 1-pcov have states -1...-g equiprobable RAS
  General model: proportion (s01/(s01+s10))*pcov covarion ON (states 1...g equiprobable), 
  proportion (s10/(s01+s10))*pcov covarion OFF (states g+1...2g equiprobable),
  proportion 1-pcov -1...-g equiprobable RAS
   */

	int i;
	signed char *P;
	
	P=seq;
	for (i=0; i<inNumSites; i++) {
		*P=SetHiddenState();
		if((covarion==Galt) | (covarion==General))
			*P=SetRASState(*P);/*will this site ever change state?*/
		if(covarion==General) {
			if((*P)>0) /*if covarion site*/
				*P=SetGeneralOnOff(*P); /*is this site initially ON (1...g) or OFF (g+1...2g)?*/
		}
		P++;
	}
}

void Huel_covarionSimulation(signed char *state, double *effectiveLength, double len)
{
/* it simulates the process ON/OFF : times when the switches occur. It stores
   the amount of time spent on "ON". */
/*This is the Huelsenbeck covarion model*/
	signed char *P;
	double r;
	double T,Ton,Ttot;

	P=state;
	Ton=Ttot=0.0; 

	while ( Ttot <= len ) {
		r=rndu();
		T=-log(r)/CovRate[*P];
		Ttot +=T;
		if (*P) Ton += T;
		if (*P) *P='\0'; else (*P)++; /* changes the state *P from ON to OFF and vice-versa*/
	}
	
	if (*P) *P='\0'; else (*P)++;;
	if (*P) *effectiveLength=Ton-(Ttot-len);
	else *effectiveLength=Ton;

}


void Galt_covarionSimulation(signed char *state, double *effectiveLength, double len)
{
/*Simulates the evolution of hidden states: generates exponential times to switches.  Equal switching rate in each hidden state, and each new state is equiprobable.  Records \sum_i r_i t_i, where r_i is rate for the ith hidden state, and t_i is the time spent in this state */
/*This is the Galtier covarion model*/
	signed char *P;
	signed char Pold;
	double r;
	double T,rsum,Ttot;

	P=state;
	Pold=*P;
	rsum=Ttot=0.0; 
	if (*P<0) {/*this is a RAS site*/
		*effectiveLength=catRate[-1*(*P)-1]*len;/*we stick with this rate category for ever*/
	}
	else {/*this is a covarion site*/
		while ( Ttot <= len ) {
			r=rndu();
			T=-log(r)/CovRate[0];/*same rate in all hidden states*/
			Ttot +=T;
		
			/*add product of rate and time to rsum*/
			rsum+=catRate[*P-1]*T;
			#ifdef DEBUG
				fprintf(stderr,"state %i, rate %f, Ttot %f, rsum %f\n",*P,catRate[*P-1],Ttot, rsum);
			#endif
		
			/*choose which state we go to*/
			Pold=*P;
			*P=SetHiddenState();
			#ifdef DEBUG
				fprintf(stderr,"new state %i\n",*P);
			#endif
		}
	
		/*the last interval: remove the */
		*P=Pold;/*reset to the state to the last one we were in before we over-ran the total edge length*/
		*effectiveLength=rsum-(Ttot-len)*catRate[*P-1];/*remove the rate*time for the part that runs beyond total edge length*/
		#ifdef DEBUG
			fprintf(stderr,"At end: state %i, rate %f, effective length %f\n",*P,catRate[*P-1], *effectiveLength);
		#endif
	}
}

void General_covarionSimulation(signed char *state, double *effectiveLength, double len)
{
/*Simulates the evolution of hidden states: generates exponential times to switches.  Chooses new states under general covarion model.    
  Records \sum_i r_i t_i, where r_i is rate for the ith hidden state, and t_i is the time spent in this state */
	signed char *P;
	signed char Pold;
	double r;
	double T,rsum,Ttot, leaveon,ponoff,pon;

	P=state;
	Pold=*P;
	rsum=Ttot=0.0; 
	if (*P<0) {/*this is a RAS site*/
		*effectiveLength=catRate[-1*(*P)-1]*len;/*we stick with this rate category for ever*/
	}
	else {/*this is a covarion site*/
		leaveon=s10+s11;/*rate of leaving an ON covarion site is s10+s11*/
		ponoff=s10/leaveon;
		pon=s01/(s01+s10);
		while ( Ttot <= len ) {
			r=rndu();
			if (*P>numCats) {/*this is an OFF state*/
				T=-log(r)/s01;/*waiting time: rate of leaving an OFF covarion site is s01*/
				Ttot +=T;/*increment total time, but rate is zero so don't change rate*time*/
				Pold=*P;
				*P-=numCats;/*we can only go to corresponding ON state*/
			}
			else {/*this is an ON state*/
				T=-log(r)/leaveon;/*waiting time*/
				Ttot +=T;/*increment total time*/
				rsum+=catRate[*P-1]*T/pon;/*add product of rate and time to rsum. rescaling by 1/pon to get correct expected edge length*/
				Pold=*P;
				if (rndu()>ponoff)
					*P=SetHiddenState();/*we go to another ON state, which may be the current one, with prob 1-ponoff*/
				else 
					*P+=numCats;/*we go to corresponding OFF state with prob ponoff*/
			}
		}
	*P=Pold;/*reset to the state to the last one we were in before we over-ran the total edge length*/
	*effectiveLength=rsum;
	if (*P<=numCats)/*we were on during the last interval*/
		 *effectiveLength-=(Ttot-len)*catRate[*P-1]/pon;/*remove the rate*time for the part that runs beyond total edge length.*/
	}
}

void MutateHiddenStates(signed char *states, double *effectiveLength, int inFromSite, int inNumSites, double len)
{
/* From site number inFromSite for inNumSites sites, it mutates the hidden states contained in
   seq (seq is changed). "len" is the length of evolution. The effective length = time spent on "ON"
   is stored in effectiveLength (each site has its own). */

	int i;
	double *L;
	/*short *S;*/
	signed char *P;
	
	P=states;
	L=effectiveLength;

	/* Here the hidden states ON /OFF are simulated for all the sites, including
	   those that will be invariable. It would speed up to skip those sites.
	   But: The variable "static short *invariable" that stores which site is
	   invariable and which is not, is a local variable of the "evolve" file.
	   Moreover, it is used for the first time (and initialized) with 
	   "EvolveSequence", that is called after the evolution of the hidden states.
	*/

	/*	if (invariableSites) {
		S = invariable + inFromSite;
		for (i=0; i<inNumSites; i++) {
			if (!(*S))
				covarionSimulation(P,L,len);
				P++;
				L++;
				S++;
		}
	} 
	else 
	*/
	{
		for (i=0; i<inNumSites; i++) {
			switch (covarion) {
				case Huel:
					Huel_covarionSimulation(P,L,len);/*Huelsenbeck model*/
				break;
				case Galt:
					Galt_covarionSimulation(P,L,len);/*Galtier model*/
				break;
				case General:
					General_covarionSimulation(P,L,len);/*General model*/
				break;
			}
			P++;
			L++;
		}
	}
}


void EvolveHiddenStateNode(TNode *anc, TNode *des, int inFromSite, int inNumSites)
{
/* recursive function. The hidden process evolves from node 'anc' to node 'des'. 
   Then the function calls itself twice to make the evolution from node 'des'
   to its 2 descendants (if it is not a tip).*/

	double len;
	
	len=des->length0;
		
	memcpy(des->hiddenStates, anc->hiddenStates, inNumSites);
	MutateHiddenStates(des->hiddenStates, des->effectiveLength0, inFromSite, inNumSites, len);
	
	if (des->tipNo==-1) {
		EvolveHiddenStateNode(des, des->branch1, inFromSite, inNumSites);
		EvolveHiddenStateNode(des, des->branch2, inFromSite, inNumSites);
	}
}


void EvolveHiddenStates(TTree *tree, int inFromSite, int inNumSites)
{
/* The evolution of the hidden process on/off is started from the root.*/
	RandomHiddenStates(tree->root->hiddenStates, inNumSites);
	switch (covarion) {
		case Huel:
			CovRate[1]=covarionRate/(2*ONfreq);	/* rate of switch when the state is ON */
			CovRate[0]=covarionRate/(2*(1-ONfreq));		/* rate when the state id OFF. */
		break;
		case Galt:
			CovRate[0]=covarionRate; /*same switching rate in all states*/
			DiscreteGamma(freqRate, catRate, gammaShape, gammaShape, numCats, 0);
		break;
		case General:
			CovRate[0]=s01; /*OFF to ON*/
			CovRate[1]=s10;/*ON to OFF*/
			CovRate[2]=s11;/*among ON states*/
			DiscreteGamma(freqRate, catRate, gammaShape, gammaShape, numCats, 0);
		break;
	}	
	EvolveHiddenStateNode(tree->root, tree->root->branch1, inFromSite, inNumSites);
	EvolveHiddenStateNode(tree->root, tree->root->branch2, inFromSite, inNumSites);
	if (!tree->rooted)
		EvolveHiddenStateNode(tree->root, tree->root->branch0, inFromSite, inNumSites);
}



void WriteHiddenStates(FILE *fv, TTree *tree, int inNumSites,TTree *nametree
)
{
	int j, n;
	signed char *P;

	fprintf(fv, "\n");
	n = numTaxa + 1;
	
	fprintf(fv, "%d\t", n);

       	switch (covarion) {
		case Huel:
	       		P=tree->root->hiddenStates;
			for (j=0; j<inNumSites; j++) {
       				 fputc(onoff[*P], fv);
       				 P++;
       			}
		break;
		case Galt: case General:
	       		P=tree->root->hiddenStates;
			for (j=0; j<inNumSites; j++) {
       				 fprintf(fv,"%i",*P);
       				 P++;
       			}
		break;
       	}

	fputc('\n', fv);
	
	if (!tree->rooted)
		WriteHiddenStatesNode(fv, tree, &n, tree->root->branch0, inNumSites,nametree);
	WriteHiddenStatesNode(fv, tree, &n, tree->root->branch1, inNumSites,nametree);
	WriteHiddenStatesNode(fv, tree, &n, tree->root->branch2, inNumSites,nametree);
}


/*MS 15/5/08: nametree is the tree where taxon names are stored*/
void WriteHiddenStatesNode(FILE *fv, TTree *tree, int *nodeNo, TNode *des, int inNumSites, TTree *nametree)
{
	int j;
	signed char *P = des->hiddenStates;
	
	if (des->tipNo==-1) {
		(*nodeNo)++;
		fprintf(fv, "%d\t", *nodeNo);

		switch (covarion) {
			case Huel:
				for (j=0; j<inNumSites; j++) {
					 fputc(onoff[*P], fv);
					 P++;
				}
			break;
			case Galt: case General:
				for (j=0; j<inNumSites; j++) {
       					 fprintf(fv,"%i",*P);
       				 	P++;
       				}
			break;
		}

		fputc('\n', fv);
		
		WriteHiddenStatesNode(fv, tree, nodeNo, des->branch1, inNumSites,nametree);
		WriteHiddenStatesNode(fv, tree, nodeNo, des->branch2, inNumSites,nametree);
	} else {
		fprintf(fv, "%s\t", nametree->names[des->tipNo]);

		switch (covarion) {
			case Huel:
				for (j=0; j<inNumSites; j++) {
					 fputc(onoff[*P], fv);
					 P++;
				}
			break;
			case Galt: case General:
				for (j=0; j<inNumSites; j++) {
       					 fprintf(fv,"%i",*P);
       				 	P++;
       				}
			break;
		}

		fputc('\n', fv);
	}
}


void WriteAncestralEffectiveLengths(FILE *fv, TTree *tree, int inNumSites, TTree *nametree)
{
	int n;
	switch(covarion) {
		case Huel:
			fprintf(fv, "\nThese effective lengths are re-scaled (by scale/(ONfreq*(1-Pinv))\n");
		break;
		case Galt:
			fprintf(fv, "\nThese effective lengths are re-scaled (by scale/(1-Pinv)\n");
		break;
		case General:
			fprintf(fv, "\nThese effective lengths are re-scaled (by scale/((1-Pinv)*(1+pcov(s01/(s01+s10)-1)))\n");
		break;
	}
	fprintf(fv, "when they are used for sequence evolution.\n");

	if (!tree->rooted)
		n = (2 * numTaxa) - 3;
	else
		n = (2 * numTaxa) - 2;
	fprintf(fv, " %d %d\n", n, inNumSites);

	n = numTaxa + 1;
	
	fprintf(fv, "%d\t", n);

	fputc('\n', fv);
	
	if (!tree->rooted)
		WriteAncestralEffLengthNode(fv, tree, &n, tree->root->branch0, inNumSites,nametree);
	WriteAncestralEffLengthNode(fv, tree, &n, tree->root->branch1, inNumSites,nametree);
	WriteAncestralEffLengthNode(fv, tree, &n, tree->root->branch2, inNumSites,nametree);
}

/*MS 15/05/08: nametree is the tree where the names are stored*/
void WriteAncestralEffLengthNode(FILE *fv, TTree *tree, int *nodeNo, TNode *des, int inNumSites, TTree *nametree)
{
	int j;
	double *P = des->effectiveLength0;
	
	if (des->tipNo==-1) {
		(*nodeNo)++;
		
		fprintf(fv, "%d\t", *nodeNo);
		for (j=0; j<inNumSites; j++) {
			fprintf(fv,"%G ",*P);
			P++;
		}

		fputc('\n', fv);
		
		WriteAncestralEffLengthNode(fv, tree, nodeNo, des->branch1, inNumSites,nametree);
		WriteAncestralEffLengthNode(fv, tree, nodeNo, des->branch2, inNumSites,nametree);
	} else {
		fprintf(fv, "%s\t", nametree->names[des->tipNo]);
		for (j=0; j<inNumSites; j++) {
			fprintf(fv,"%G ",*P);
			P++;
		}

		fputc('\n', fv);
	}

}

